@@ -30,7 +30,8 @@ import mesonbuild.mlog
 import mesonbuild.compilers
 import mesonbuild.environment
 import mesonbuild.mesonlib
-from mesonbuild.mesonlib import is_windows, is_osx, is_cygwin
+import mesonbuild.coredata
+from mesonbuild.mesonlib import is_linux, is_windows, is_osx, is_cygwin, windows_proof_rmtree
 from mesonbuild.environment import Environment
 from mesonbuild.dependencies import DependencyException
 from mesonbuild.dependencies import PkgConfigDependency, ExternalProgram
@@ -38,6 +39,7 @@ from mesonbuild.dependencies import PkgConfigDependency, ExternalProgram
 from run_tests import exe_suffix, get_fake_options, FakeEnvironment
 from run_tests import get_builddir_target_args, get_backend_commands, Backend
 from run_tests import ensure_backend_detects_changes, run_configure_inprocess
+from run_tests import should_run_linux_cross_tests
 
 
 def get_dynamic_section_entry(fname, entry):
@@ -109,7 +111,7 @@ class InternalTests(unittest.TestCase):
 
     def test_compiler_args_class(self):
         cargsfunc = mesonbuild.compilers.CompilerArgs
-        c = mesonbuild.environment.CCompiler([], 'fake', False)
+        c = mesonbuild.compilers.CCompiler([], 'fake', False)
         # Test that bad initialization fails
         self.assertRaises(TypeError, cargsfunc, [])
         self.assertRaises(TypeError, cargsfunc, [], [])
@@ -416,9 +418,9 @@ class BasePlatformTests(unittest.TestCase):
         self.backend = getattr(Backend, os.environ.get('MESON_UNIT_TEST_BACKEND', 'ninja'))
         self.meson_args = [os.path.join(src_root, 'meson.py'), '--backend=' + self.backend.name]
         self.meson_command = [sys.executable] + self.meson_args
-        self.mconf_command = [sys.executable, os.path.join(src_root, 'mesonconf.py')]
-        self.mintro_command = [sys.executable, os.path.join(src_root, 'mesonintrospect.py')]
-        self.mtest_command = [sys.executable, os.path.join(src_root, 'mesontest.py'), '-C', self.builddir]
+        self.mconf_command = [sys.executable, os.path.join(src_root, 'meson.py'), 'configure']
+        self.mintro_command = [sys.executable, os.path.join(src_root, 'meson.py'), 'introspect']
+        self.mtest_command = [sys.executable, os.path.join(src_root, 'meson.py'), 'test', '-C', self.builddir]
         # Backend-specific build commands
         self.build_command, self.clean_command, self.test_command, self.install_command, \
             self.uninstall_command = get_backend_commands(self.backend)
@@ -445,7 +447,7 @@ class BasePlatformTests(unittest.TestCase):
             print(f.read())
 
     def tearDown(self):
-        shutil.rmtree(self.builddir)
+        windows_proof_rmtree(self.builddir)
         os.environ = self.orig_env
         super().tearDown()
 
@@ -466,7 +468,7 @@ class BasePlatformTests(unittest.TestCase):
         return output
 
     def init(self, srcdir, extra_args=None, default_args=True, inprocess=False):
-        self.assertTrue(os.path.exists(srcdir))
+        self.assertPathExists(srcdir)
         if extra_args is None:
             extra_args = []
         if not isinstance(extra_args, list):
@@ -534,7 +536,7 @@ class BasePlatformTests(unittest.TestCase):
         self._run(self.mconf_command + [arg, self.builddir])
 
     def wipe(self):
-        shutil.rmtree(self.builddir)
+        windows_proof_rmtree(self.builddir)
 
     def utime(self, f):
         ensure_backend_detects_changes(self.backend)
@@ -627,6 +629,14 @@ class BasePlatformTests(unittest.TestCase):
         else:
             raise RuntimeError('Invalid backend: {!r}'.format(self.backend.name))
 
+    def assertPathExists(self, path):
+        m = 'Path {!r} should exist'.format(path)
+        self.assertTrue(os.path.exists(path), msg=m)
+
+    def assertPathDoesNotExist(self, path):
+        m = 'Path {!r} should not exist'.format(path)
+        self.assertFalse(os.path.exists(path), msg=m)
+
 
 class AllPlatformTests(BasePlatformTests):
     '''
@@ -701,7 +711,7 @@ class AllPlatformTests(BasePlatformTests):
         static_linker = env.detect_static_linker(cc)
         if is_windows():
             raise unittest.SkipTest('https://github.com/mesonbuild/meson/issues/1526')
-        if not isinstance(static_linker, mesonbuild.compilers.ArLinker):
+        if not isinstance(static_linker, mesonbuild.linkers.ArLinker):
             raise unittest.SkipTest('static linker is not `ar`')
         # Configure
         self.init(testdir)
@@ -772,11 +782,11 @@ class AllPlatformTests(BasePlatformTests):
         exename = os.path.join(self.installdir, 'usr/bin/prog' + exe_suffix)
         testdir = os.path.join(self.common_test_dir, '8 install')
         self.init(testdir)
-        self.assertFalse(os.path.exists(exename))
+        self.assertPathDoesNotExist(exename)
         self.install()
-        self.assertTrue(os.path.exists(exename))
+        self.assertPathExists(exename)
         self.uninstall()
-        self.assertFalse(os.path.exists(exename))
+        self.assertPathDoesNotExist(exename)
 
     def test_testsetups(self):
         if not shutil.which('valgrind'):
@@ -867,10 +877,10 @@ class AllPlatformTests(BasePlatformTests):
         self.build()
         genfile = os.path.join(self.builddir, 'generated.dat')
         exe = os.path.join(self.builddir, 'fooprog' + exe_suffix)
-        self.assertTrue(os.path.exists(genfile))
-        self.assertFalse(os.path.exists(exe))
+        self.assertPathExists(genfile)
+        self.assertPathDoesNotExist(exe)
         self.build(target=('fooprog' + exe_suffix))
-        self.assertTrue(os.path.exists(exe))
+        self.assertPathExists(exe)
 
     def test_internal_include_order(self):
         testdir = os.path.join(self.common_test_dir, '138 include order')
@@ -936,8 +946,8 @@ class AllPlatformTests(BasePlatformTests):
         clang = mesonbuild.compilers.ClangCompiler
         intel = mesonbuild.compilers.IntelCompiler
         msvc = mesonbuild.compilers.VisualStudioCCompiler
-        ar = mesonbuild.compilers.ArLinker
-        lib = mesonbuild.compilers.VisualStudioLinker
+        ar = mesonbuild.linkers.ArLinker
+        lib = mesonbuild.linkers.VisualStudioLinker
         langs = [('c', 'CC'), ('cpp', 'CXX')]
         if not is_windows():
             langs += [('objc', 'OBJC'), ('objcpp', 'OBJCXX')]
@@ -948,6 +958,7 @@ class AllPlatformTests(BasePlatformTests):
             # Detect with evar and do sanity checks on that
             if evar in os.environ:
                 ecc = getattr(env, 'detect_{}_compiler'.format(lang))(False)
+                self.assertTrue(ecc.version)
                 elinker = env.detect_static_linker(ecc)
                 # Pop it so we don't use it for the next detection
                 evalue = os.environ.pop(evar)
@@ -971,6 +982,7 @@ class AllPlatformTests(BasePlatformTests):
                 self.assertEqual(ecc.get_exelist(), shlex.split(evalue))
             # Do auto-detection of compiler based on platform, PATH, etc.
             cc = getattr(env, 'detect_{}_compiler'.format(lang))(False)
+            self.assertTrue(cc.version)
             linker = env.detect_static_linker(cc)
             # Check compiler type
             if isinstance(cc, gnu):
@@ -1004,11 +1016,18 @@ class AllPlatformTests(BasePlatformTests):
                 self.assertTrue(is_windows())
                 self.assertIsInstance(linker, lib)
                 self.assertEqual(cc.id, 'msvc')
+                self.assertTrue(hasattr(cc, 'is_64'))
+                # If we're in the appveyor CI, we know what the compiler will be
+                if 'arch' in os.environ:
+                    if os.environ['arch'] == 'x64':
+                        self.assertTrue(cc.is_64)
+                    else:
+                        self.assertFalse(cc.is_64)
             # Set evar ourselves to a wrapper script that just calls the same
             # exelist + some argument. This is meant to test that setting
             # something like `ccache gcc -pipe` or `distcc ccache gcc` works.
             wrapper = os.path.join(testdir, 'compiler wrapper.py')
-            wrappercc = [sys.executable, wrapper] + cc.get_exelist() + cc.get_always_args()
+            wrappercc = [sys.executable, wrapper] + cc.get_exelist() + ['-DSOME_ARG']
             wrappercc_s = ''
             for w in wrappercc:
                 wrappercc_s += shlex.quote(w) + ' '
@@ -1027,6 +1046,10 @@ class AllPlatformTests(BasePlatformTests):
             # Ensure that the exelist is correct
             self.assertEqual(wcc.get_exelist(), wrappercc)
             self.assertEqual(wlinker.get_exelist(), wrapperlinker)
+            # Ensure that the version detection worked correctly
+            self.assertEqual(cc.version, wcc.version)
+            if hasattr(cc, 'is_64'):
+                self.assertEqual(cc.is_64, wcc.is_64)
 
     def test_always_prefer_c_compiler_for_asm(self):
         testdir = os.path.join(self.common_test_dir, '141 c cpp and asm')
@@ -1161,18 +1184,50 @@ class AllPlatformTests(BasePlatformTests):
         self.build()
         self.run_tests()
 
-    def test_dist(self):
+    def test_dist_git(self):
         if not shutil.which('git'):
             raise unittest.SkipTest('Git not found')
+
+        def git_init(project_dir):
+            subprocess.check_call(['git', 'init'], cwd=project_dir)
+            subprocess.check_call(['git', 'config',
+                                   'user.name', 'Author Person'], cwd=project_dir)
+            subprocess.check_call(['git', 'config',
+                                   'user.email', 'teh_coderz@example.com'], cwd=project_dir)
+            subprocess.check_call(['git', 'add', 'meson.build', 'distexe.c'], cwd=project_dir)
+            subprocess.check_call(['git', 'commit', '-a', '-m', 'I am a project'], cwd=project_dir)
+
         try:
-            self.dist_impl()
+            self.dist_impl(git_init)
         except PermissionError:
             # When run under Windows CI, something (virus scanner?)
             # holds on to the git files so cleaning up the dir
             # fails sometimes.
             pass
 
-    def dist_impl(self):
+    def test_dist_hg(self):
+        if not shutil.which('hg'):
+            raise unittest.SkipTest('Mercurial not found')
+        if self.backend is not Backend.ninja:
+            raise unittest.SkipTest('Dist is only supported with Ninja')
+
+        def hg_init(project_dir):
+            subprocess.check_call(['hg', 'init'], cwd=project_dir)
+            with open(os.path.join(project_dir, '.hg', 'hgrc'), 'w') as f:
+                print('[ui]', file=f)
+                print('username=Author Person <teh_coderz@example.com>', file=f)
+            subprocess.check_call(['hg', 'add', 'meson.build', 'distexe.c'], cwd=project_dir)
+            subprocess.check_call(['hg', 'commit', '-m', 'I am a project'], cwd=project_dir)
+
+        try:
+            self.dist_impl(hg_init)
+        except PermissionError:
+            # When run under Windows CI, something (virus scanner?)
+            # holds on to the hg files so cleaning up the dir
+            # fails sometimes.
+            pass
+
+    def dist_impl(self, vcs_init):
         # Create this on the fly because having rogue .git directories inside
         # the source tree leads to all kinds of trouble.
         with tempfile.TemporaryDirectory() as project_dir:
@@ -1189,19 +1244,13 @@ int main(int argc, char **argv) {
     return 0;
 }
 ''')
-            subprocess.check_call(['git', 'init'], cwd=project_dir)
-            subprocess.check_call(['git', 'config',
-                                   'user.name', 'Author Person'], cwd=project_dir)
-            subprocess.check_call(['git', 'config',
-                                   'user.email', 'teh_coderz@example.com'], cwd=project_dir)
-            subprocess.check_call(['git', 'add', 'meson.build', 'distexe.c'], cwd=project_dir)
-            subprocess.check_call(['git', 'commit', '-a', '-m', 'I am a project'], cwd=project_dir)
+            vcs_init(project_dir)
             self.init(project_dir)
             self.build('dist')
             distfile = os.path.join(self.distdir, 'disttest-1.4.3.tar.xz')
             checksumfile = distfile + '.sha256sum'
-            self.assertTrue(os.path.exists(distfile))
-            self.assertTrue(os.path.exists(checksumfile))
+            self.assertPathExists(distfile)
+            self.assertPathExists(checksumfile)
 
     def test_rpath_uses_ORIGIN(self):
         '''
@@ -1221,6 +1270,27 @@ int main(int argc, char **argv) {
             for path in rpath.split(':'):
                 self.assertTrue(path.startswith('$ORIGIN'), msg=(each, path))
 
+    def test_dash_d_dedup(self):
+        testdir = os.path.join(self.unit_test_dir, '10 d dedup')
+        self.init(testdir)
+        cmd = self.get_compdb()[0]['command']
+        self.assertTrue('-D FOO -D BAR' in cmd or
+                        '"-D" "FOO" "-D" "BAR"' in cmd or
+                        '/D FOO /D BAR' in cmd or
+                        '"/D" "FOO" "/D" "BAR"' in cmd)
+
+    def test_all_forbidden_targets_tested(self):
+        '''
+        Test that all forbidden targets are tested in the '159 reserved targets'
+        test. Needs to be a unit test because it accesses Meson internals.
+        '''
+        testdir = os.path.join(self.common_test_dir, '159 reserved targets')
+        targets = mesonbuild.coredata.forbidden_target_names
+        # We don't actually define a target with this name
+        targets.pop('build.ninja')
+        for i in targets:
+            self.assertPathExists(os.path.join(testdir, i))
+
 
 class FailureTests(BasePlatformTests):
     '''
@@ -1239,7 +1309,7 @@ class FailureTests(BasePlatformTests):
 
     def tearDown(self):
         super().tearDown()
-        shutil.rmtree(self.srcdir)
+        windows_proof_rmtree(self.srcdir)
 
     def assertMesonRaises(self, contents, match, extra_args=None, langs=None):
         '''
@@ -1383,6 +1453,23 @@ class WindowsTests(BasePlatformTests):
         self.assertPathEqual(prog.get_command()[0], sys.executable)
         self.assertPathBasenameEqual(prog.get_path(), 'test-script-ext.py')
 
+    def test_ignore_libs(self):
+        '''
+        Test that find_library on libs that are to be ignored returns an empty
+        array of arguments. Must be a unit test because we cannot inspect
+        ExternalLibraryHolder from build files.
+        '''
+        testdir = os.path.join(self.platform_test_dir, '1 basic')
+        env = Environment(testdir, self.builddir, self.meson_command,
+                          get_fake_options(self.prefix), [])
+        cc = env.detect_c_compiler(False)
+        if cc.id != 'msvc':
+            raise unittest.SkipTest('Not using MSVC')
+        # To force people to update this test, and also test
+        self.assertEqual(set(cc.ignore_libs), {'c', 'm', 'pthread'})
+        for l in cc.ignore_libs:
+            self.assertEqual(cc.find_library(l, env, []), [])
+
 
 class LinuxlikeTests(BasePlatformTests):
     '''
@@ -1538,35 +1625,35 @@ class LinuxlikeTests(BasePlatformTests):
 
         # File without aliases set.
         nover = os.path.join(libpath, 'libnover.so')
-        self.assertTrue(os.path.exists(nover))
+        self.assertPathExists(nover)
         self.assertFalse(os.path.islink(nover))
         self.assertEqual(get_soname(nover), 'libnover.so')
         self.assertEqual(len(glob(nover[:-3] + '*')), 1)
 
         # File with version set
         verset = os.path.join(libpath, 'libverset.so')
-        self.assertTrue(os.path.exists(verset + '.4.5.6'))
+        self.assertPathExists(verset + '.4.5.6')
         self.assertEqual(os.readlink(verset), 'libverset.so.4')
         self.assertEqual(get_soname(verset), 'libverset.so.4')
         self.assertEqual(len(glob(verset[:-3] + '*')), 3)
 
         # File with soversion set
         soverset = os.path.join(libpath, 'libsoverset.so')
-        self.assertTrue(os.path.exists(soverset + '.1.2.3'))
+        self.assertPathExists(soverset + '.1.2.3')
         self.assertEqual(os.readlink(soverset), 'libsoverset.so.1.2.3')
         self.assertEqual(get_soname(soverset), 'libsoverset.so.1.2.3')
         self.assertEqual(len(glob(soverset[:-3] + '*')), 2)
 
         # File with version and soversion set to same values
         settosame = os.path.join(libpath, 'libsettosame.so')
-        self.assertTrue(os.path.exists(settosame + '.7.8.9'))
+        self.assertPathExists(settosame + '.7.8.9')
         self.assertEqual(os.readlink(settosame), 'libsettosame.so.7.8.9')
         self.assertEqual(get_soname(settosame), 'libsettosame.so.7.8.9')
         self.assertEqual(len(glob(settosame[:-3] + '*')), 2)
 
         # File with version and soversion set to different values
         bothset = os.path.join(libpath, 'libbothset.so')
-        self.assertTrue(os.path.exists(bothset + '.1.2.3'))
+        self.assertPathExists(bothset + '.1.2.3')
         self.assertEqual(os.readlink(bothset), 'libbothset.so.1.2.3')
         self.assertEqual(os.readlink(bothset + '.1.2.3'), 'libbothset.so.4.5.6')
         self.assertEqual(get_soname(bothset), 'libbothset.so.1.2.3')
@@ -1653,9 +1740,9 @@ class LinuxlikeTests(BasePlatformTests):
     def test_unity_subproj(self):
         testdir = os.path.join(self.common_test_dir, '49 subproject')
         self.init(testdir, extra_args='--unity=subprojects')
-        self.assertTrue(os.path.exists(os.path.join(self.builddir, 'subprojects/sublib/simpletest@exe/simpletest-unity.c')))
-        self.assertTrue(os.path.exists(os.path.join(self.builddir, 'subprojects/sublib/sublib@sha/sublib-unity.c')))
-        self.assertFalse(os.path.exists(os.path.join(self.builddir, 'user@exe/user-unity.c')))
+        self.assertPathExists(os.path.join(self.builddir, 'subprojects/sublib/simpletest@exe/simpletest-unity.c'))
+        self.assertPathExists(os.path.join(self.builddir, 'subprojects/sublib/sublib@sha/sublib-unity.c'))
+        self.assertPathDoesNotExist(os.path.join(self.builddir, 'user@exe/user-unity.c'))
         self.build()
 
     def test_installed_modes(self):
@@ -1761,19 +1848,65 @@ class LinuxlikeTests(BasePlatformTests):
         self.init(testdir)
         # NOTE: .pc file has -Lfoo -lfoo -Lbar -lbar but pkg-config reorders
         # the flags before returning them to -Lfoo -Lbar -lfoo -lbar
-        expected_order = ['-L/me/first', '-L/me/second','-lfoo1', '-lfoo2',
-                          '-L/me/third', '-L/me/fourth', '-lfoo3', '-lfoo4']
+        # but pkgconf seems to not do that. Sigh. Support both.
+        expected_order = [('-L/me/first', '-lfoo1'),
+                          ('-L/me/second', '-lfoo2'),
+                          ('-L/me/first', '-L/me/second'),
+                          ('-lfoo1', '-lfoo2'),
+                          ('-L/me/second', '-L/me/third'),
+                          ('-L/me/third', '-L/me/fourth',),
+                          ('-L/me/third', '-lfoo3'),
+                          ('-L/me/fourth', '-lfoo4'),
+                          ('-lfoo3', '-lfoo4'),
+                          ]
         with open(os.path.join(self.builddir, 'build.ninja')) as ifile:
             for line in ifile:
-                if expected_order[0] in line:
-                    previous_index = line.index(expected_order[0])
-                    for entry in expected_order[1:]:
-                        current_index = line.index(entry)
-                        self.assertLess(previous_index, current_index)
-                        previous_index = current_index
+                if expected_order[0][0] in line:
+                    for first, second in expected_order:
+                        self.assertLess(line.index(first), line.index(second))
                     return
         raise RuntimeError('Linker entries not found in the Ninja file.')
 
+    def test_introspect_dependencies(self):
+        '''
+        Tests that mesonintrospect --dependencies returns expected output.
+        '''
+        testdir = os.path.join(self.framework_test_dir, '7 gnome')
+        self.init(testdir)
+        glib_found = False
+        gobject_found = False
+        deps = self.introspect('--dependencies')
+        self.assertIsInstance(deps, list)
+        for dep in deps:
+            self.assertIsInstance(dep, dict)
+            self.assertIn('name', dep)
+            self.assertIn('compile_args', dep)
+            self.assertIn('link_args', dep)
+            if dep['name'] == 'glib-2.0':
+                glib_found = True
+            elif dep['name'] == 'gobject-2.0':
+                gobject_found = True
+        self.assertTrue(glib_found)
+        self.assertTrue(gobject_found)
+
+    def test_build_rpath(self):
+        testdir = os.path.join(self.unit_test_dir, '11 build_rpath')
+        self.init(testdir)
+        self.build()
+        build_rpath = get_rpath(os.path.join(self.builddir, 'prog'))
+        self.assertEqual(build_rpath, '$ORIGIN/sub:/foo/bar')
+        self.install()
+        install_rpath = get_rpath(os.path.join(self.installdir, 'usr/bin/prog'))
+        self.assertEqual(install_rpath, '/baz')
+
+    def test_pch_with_address_sanitizer(self):
+        testdir = os.path.join(self.common_test_dir, '13 pch')
+        self.init(testdir, ['-Db_sanitize=address'])
+        self.build()
+        compdb = self.get_compdb()
+        for i in compdb:
+            self.assertIn("-fsanitize=address", i["command"])
+
 class LinuxArmCrossCompileTests(BasePlatformTests):
     '''
     Tests that verify cross-compilation to Linux/ARM
@@ -1807,7 +1940,7 @@ class RewriterTests(unittest.TestCase):
         self.test_dir = os.path.join(src_root, 'test cases/rewrite')
 
     def tearDown(self):
-        shutil.rmtree(self.tmpdir)
+        windows_proof_rmtree(self.tmpdir)
 
     def read_contents(self, fname):
         with open(os.path.join(self.workdir, fname)) as f:
@@ -1869,4 +2002,12 @@ def unset_envs():
 
 if __name__ == '__main__':
     unset_envs()
-    unittest.main(buffer=True)
+    cases = ['InternalTests', 'AllPlatformTests', 'FailureTests']
+    if is_linux():
+        cases += ['LinuxlikeTests']
+        if should_run_linux_cross_tests():
+            cases += ['LinuxArmCrossCompileTests']
+    elif is_windows():
+        cases += ['WindowsTests']
+
+    unittest.main(defaultTest=cases, buffer=True)
